/*
 * bootstrap.c -- This program loads the Linux/68k kernel into an Amiga
 *                and and launches it.
 *
 * Copyright (C) 1994 by Waldorf GMBH,
 * written by Ralf Baechle
 *
 * Loosly based on bootstrap.c for Linux/68k
 *
 * Copyright (C) 1993 by Hamish Macdonald, Greg Harp
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 */

/*
 * Defaults, may be overiden by option or environment.
 */
static char *kernel_name = "a:vmlinux";
static char *ramdisk_name = NULL;

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/file.h>
#include <sys/types.h>
#include <unistd.h>

/* Amiga bootstrap include file */
#include "bootstrap.h"

/*
 * required Linux/MIPS include files
 */
#include <linux/a.out.h>
#include <linux/bootinfo.h>

extern char *optarg;

struct exec kexec;
char *memptr;
u_long start_mem;
u_long mem_size;
u_long rd_size;

struct bootinfo bi;

void usage(void)
{
	fprintf (stderr, "Usage:\n"
		 "\tbootstrap [-d] [-k kernel_executable] [-r ramdisk_file]"
		 " [option...]\n");
	exit (EXIT_FAILURE);
}

extern char copyall, copyallend;

int main(int argc, char *argv[])
{
	int ch, debugflag = 0, kfd, rfd = -1, i;
	long fast_total = 0;	     /* total Fast RAM in system */
	struct MemHeader *mnp;
	struct ConfigDev *cdp = NULL;
	u_long memreq;
	struct nlist *nl;
	void (*startfunc)(void);
	long startcodesize;
	u_long *stack, kbi_offset;

	/* print the greet message */
	puts("Linux/MIPS ARC Bootstrap version 0.0");

	/*
	 * Machine is Deskstation Tyne
	 */
	bi.machtype = MACH_TYNE;

	/* check arguments */
	while ((ch = getopt(argc, argv, "dk:r:")) != EOF)
		switch (ch) {
		    case 'd':
			debugflag = 1;
			break;
		    case 'k':
			kernel_name = optarg;
			break;
		    case 'r':
			ramdisk_name = optarg;
			break;
		    case '?':
		    default:
			usage();
		}
	argc -= optind;
	argv += optind;


    /*
     * FIXME: should use the ARC BIOS
     */
    bi.model      = DESKSTATION_TYNE;
    bi.cputype    = R4600;
    bi.memlower   = 0x80000000;
    bi.memupper   = 0x80800000;
    bi.sec_cache  = 512k;
    bi.dma_cache  = 512k;

    /*
     * tell us where the kernel will go
     */
    printf("\nThe kernel will be located at %08lx\n", start_mem);

    /*
     * verify that there is enough RAM
     */
    if(bi.memupper - bi.memlower < 0x800000) {
        fprintf(stderr, "Not enough RAM, aborting\n\r");
        exit(5);
        }

    /*
     * open kernel executable and read exec header
     */
        if ((kfd = open (kernel_name, O_RDONLY)) == -1) {
        fprintf (stderr, "Unable to open kernel file %s\n", kernel_name);
        exit (EXIT_FAILURE);
        }

    if (read (kfd, (void *)&kexec, sizeof(kexec)) != sizeof(kexec)) {
        fprintf (stderr, "Unable to read exec header from %s\n", kernel_name);
        exit (EXIT_FAILURE);
        }

    /*
     * No RAM-disk image support yet.
     */
    if (ramdisk_name) {
        if ((rfd = open (ramdisk_name, O_RDONLY)) == -1) {
            fprintf (stderr, "Unable to open ramdisk file %s\n", ramdisk_name);
            exit (EXIT_FAILURE);
            }
	/*
         * record ramdisk size
         */
        bi.ramdisk_size = (lseek (rfd, 0, L_XTND) + 1023) >> 10;
	}
    else
        bi.ramdisk_size = 0;

    rd_size = bi.ramdisk_size << 10;
    bi.ramdisk_addr = (u_long)start_mem + mem_size - rd_size;

    /*
     * find offset to boot_info structure
     */
    if (!(nl = get_nlist (kernel_name, "_boot_info"))) {
        perror ("get_nlist");
        exit (EXIT_FAILURE);
    } else {
        kbi_offset = nl->n_value - kexec.a_entry;
        free (nl);
        }

    memreq = kexec.a_text + kexec.a_data + rd_size;
    if (!(memptr = (char *) malloc(memreq))) {
        fprintf (stderr, "Unable to allocate memory\n");
        exit (EXIT_FAILURE);
	}
    memset(memptr, 0, memreq);

    if (lseek (kfd, N_TXTOFF(kexec), L_SET) == -1) {
        fprintf (stderr, "Failed to seek to text\n");
        free(memptr);
        exit (EXIT_FAILURE);
        }
    if (read (kfd, memptr, kexec.a_text) != kexec.a_text) {
        fprintf (stderr, "Failed to read text\n");
        free(memptr);
        exit (EXIT_FAILURE);
	}
    if (lseek (kfd, N_DATOFF(kexec), L_SET) == -1) {
        fprintf (stderr, "Failed to seek to data\n");
        free(memptr);
        exit (EXIT_FAILURE);
	}
    if (read (kfd, memptr + kexec.a_text, kexec.a_data) != kexec.a_data) {
        fprintf (stderr, "Failed to read data\n");
        free(memptr);
        exit (EXIT_FAILURE);
	}
    close (kfd);

    /*
     * copy the boot_info struct to the kernel image
     */
    memcpy ((void *)(memptr + kbi_offset), &bi, sizeof(bi));

    if (rfd != -1) {
        if (lseek (rfd, 0, L_SET) == -1) {
            fprintf (stderr, "Failed to seek to beginning of ramdisk file\n");
            free(memptr);
            exit (EXIT_FAILURE);
            }
        if (read (rfd, memptr + kexec.a_text + kexec.a_data,
            rd_size) != rd_size) {
            fprintf (stderr, "Failed to read ramdisk file\n");
            free(memptr);
            exit (EXIT_FAILURE);
            }
        close (rfd);
    }

    /*
     * Should check whether Kernel & RAMdisk moving doesn't overwrite
     * the bootstraper during startup.
     */
    if (debugflag) {
        if (bi.ramdisk_size)
            printf ("RAM disk at %#lx, size is %ldK\n",
                    (u_long)memptr + kexec.a_text + kexec.a_data,
                    bi.ramdisk_size);

        printf ("\nKernel text at %#lx, code size %d\n",
                start_mem + N_TXTADDR(kexec), kexec.a_text);
        printf ("Kernel data at %#lx, data size %d\n",
                start_mem + N_DATADDR(kexec), kexec.a_data );
        printf ("Kernel bss  at %#lx, bss  size %d\n",
                start_mem + N_BSSADDR(kexec), kexec.a_bss );

        printf ("\nKernel entry is %#x\n", kexec.a_entry );

        printf ("ramdisk dest top is %#lx\n", start_mem + mem_size);
        printf ("ramdisk lower limit is %#lx\n",
                (u_long)(memptr + kexec.a_text + kexec.a_data));
        printf ("ramdisk src top is %#lx\n",
                (u_long)(memptr + kexec.a_text + kexec.a_data)
                + rd_size);

        printf ("Type a key to continue the Linux boot...");
            fflush (stdout);
            getchar();
        }

	/*
         * wait for things to settle down
         */
	sleep(2);

	sti()

        /*
         * execute the copy-and-go code
         */
	startfunc();

	/* NOTREACHED */
}
